# -*- coding: utf-8 -*-
"""

"""
from typing import Optional

from numpy import array, delete, dot, logical_and, ndarray, in1d, all, sqrt, zeros
from numpy.linalg import det

from pyfem.bc.BaseBC import BaseBC
from pyfem.io.Amplitude import Amplitude
from pyfem.io.BC import BC
from pyfem.io.Dof import Dof
from pyfem.io.Solver import Solver
from pyfem.isoelements.IsoElementShape import iso_element_shape_dict
from pyfem.isoelements.get_iso_element_type import get_iso_element_type
from pyfem.mesh.MeshData import MeshData
from pyfem.utils.colors import error_style


class NeumannBCDistributed(BaseBC):
    r"""
    **Neumann边界条件：分布载荷**

    基于边界条件的属性、自由度属性、网格对象、求解器属性和幅值属性获取系统线性方程组 :math:`{\mathbf{K u}} = {\mathbf{f}}` 中对应等式右边项 :math:`{\mathbf{f}}` 的约束信息。

    Neumann分布载荷边界条件只能施加于边界表面列表 :py:attr:`bc_surface`，其中边界表面列表是由元组（单元编号，单元面名称）对象组成的列表。

    边界表面列表 :py:attr:`bc_surface` 可以由边界条件属性中的节点集合 :py:attr:`pyfem.io.BC.BC.node_sets` 和单元集合 :py:attr:`pyfem.io.BC.BC.element_sets` 通过函数 :py:meth:`get_surface_from_elements_nodes` 确定，也可以由边界条件属性中的边界单元集合 :py:attr:`pyfem.io.BC.BC.bc_element_sets` 通过函数 :py:meth:`get_surface_from_bc_element` 确定。

    对象创建时更新自由度序号列表 :py:attr:`bc_node_ids` 和对应等式右边项取值列表 :py:attr:`bc_fext` 。

    ========================================
    理论-基于等参元的计算方法
    ========================================

    ----------------------------------------
    1. 基于变分法的基本公式
    ----------------------------------------

    考虑在单元边界上施加某个分布载荷的情况，对于二维问题，面力 :math:`\mathbf{\bar p}` 在 :math:`\left( m \right)` 号单元所产生的等效节点载荷为：

    .. math::
        {{\mathbf{R}}^{\left( m \right)}} = \int\limits_{s_p^{\left( m \right)}} {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}{\text{d}}s}

    对于三维问题，面力 :math:`\mathbf{\bar p}` 在 :math:`\left( m \right)` 号单元所产生的等效节点载荷为：

    .. math::
        {{\mathbf{R}}^{\left( m \right)}} = \iint\limits_{S_p^{\left( m \right)}} {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}{\text{d}}S}

    因此，对于边界上的分布载荷，我们需要处理单元边界上的第一类曲线积分和第一类曲面积分。

    ----------------------------------------
    2. 二维单元边界的第一类曲线积分
    ----------------------------------------

    通过坐标变换，我们可以建立全局坐标和局部坐标的关系，二维下的弧微分可以表示为：

    .. math::
        {\text{d}}s = \sqrt {{{\left( {{\text{d}}x} \right)}^2} + {{\left( {{\text{d}}y} \right)}^2}}  = \sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}{\text{d}}\xi  + \frac{{\partial x}}{{\partial \eta }}{\text{d}}\eta } \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}{\text{d}}\xi  + \frac{{\partial y}}{{\partial \eta }}{\text{d}}\eta } \right)}^2}}


    **2.1 以四节点四边形等参元为例子** ::

      (-1,1)           (1,1)
        3---------------2
        |       η       |
        |       |       |
        |       o--ξ    |
        |               |
        |               |
        0---------------1
      (-1,-1)          (1,-1)

    边界上的第一类曲线积分可以表示为：

    .. math::
        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\xi  =  \pm 1}} = \int_{ - 1}^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}\sqrt {{{\left( {\frac{{\partial x}}{{\partial \eta }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \eta }}} \right)}^2}} {\text{d}}\eta }

    .. math::
        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\eta  =  \pm 1}} = \int_{ - 1}^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}\sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}} \right)}^2}} {\text{d}}\xi }

    **2.2 以三节点三角形等参元为例子** ::

       (0,1)
        2
        * *
        *   *
        *     *
        η       *
        |         *
        0---ξ * * * 1
       (0,0)       (1,0)

    边界上的第一类曲线积分可以表示为：

    .. math::
        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\xi  = 0}} = \int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}\sqrt {{{\left( {\frac{{\partial x}}{{\partial \eta }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \eta }}} \right)}^2}} {\text{d}}\eta }

        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\eta  = 0}} = \int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}\sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}} \right)}^2}} {\text{d}}\xi }

    注意，对于斜边 1-2，我们有直线方程 :math:`\eta  = 1 - \xi` ，此时 :math:`{\text{d}}\eta = -{\text{d}}\xi` ：

    .. math::
        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\eta  = 1 - \xi }} = \int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}\sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }} - \frac{{\partial x}}{{\partial \eta }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }} - \frac{{\partial y}}{{\partial \eta }}} \right)}^2}} {\text{d}}\xi }

    但是这种表达形式难以用于三维问题，因此我们对考虑将斜边 1-2 投影到其他坐标轴方向，由于斜边 1-2 的直线方程是已知的，因此有：

    .. math::
        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\eta  = 1 - \xi }} = \int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}\sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}} \right)}^2}} \sqrt {1 + {{\left( {\frac{{\partial \eta }}{{\partial \xi }}} \right)}^2}} {\text{d}}\xi }  = \int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}\sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}} \right)}^2}} \sqrt 2 {\text{d}}\xi }

    通过选取合适的积分点，采用数值积分即可求得积分值。

    ----------------------------------------
    3. 三维单元边界的第一类曲面积分
    ----------------------------------------

    对于三维问题，我们有：

    .. math::
        {\text{d}}\vec \xi  = \frac{{\partial x}}{{\partial \xi }}{\text{d}}\xi \vec i + \frac{{\partial y}}{{\partial \xi }}{\text{d}}\xi \vec j + \frac{{\partial z}}{{\partial \xi }}{\text{d}}\xi \vec k

        {\text{d}}\vec \eta  = \frac{{\partial x}}{{\partial \eta }}{\text{d}}\eta \vec i + \frac{{\partial y}}{{\partial \eta }}{\text{d}}\eta \vec j + \frac{{\partial z}}{{\partial \eta }}{\text{d}}\eta \vec k

        {\text{d}}\vec \zeta  = \frac{{\partial x}}{{\partial \zeta }}{\text{d}}\zeta \vec i + \frac{{\partial y}}{{\partial \zeta }}{\text{d}}\zeta \vec j + \frac{{\partial z}}{{\partial \zeta }}{\text{d}}\zeta \vec k

    此时体积微元的变换可以表示为：

    .. math::
        {\text{d}}V = \left( {{\text{d}}\vec \xi  \times {\text{d}}\vec \eta } \right) \cdot {\text{d}}\vec \zeta  = \det \left( {\mathbf{J}} \right){\text{d}}\xi {\text{d}}\eta {\text{d}}\zeta

    由 :math:`{\text{d}}\vec \xi` 和 :math:`{\text{d}}\vec \eta` 组成的面积微元矢量 :math:`{\text{d}}\vec S` 可以表示为：

    .. math::
        {\text{d}}\vec S = {\text{d}}\vec \xi  \times {\text{d}}\vec \eta  = \det \left( {\begin{array}{*{20}{c}}
          {\vec i}&{\vec j}&{\vec k} \\
          {\frac{{\partial x}}{{\partial \xi }}}&{\frac{{\partial y}}{{\partial \xi }}}&{\frac{{\partial z}}{{\partial \xi }}} \\
          {\frac{{\partial x}}{{\partial \eta }}}&{\frac{{\partial y}}{{\partial \eta }}}&{\frac{{\partial z}}{{\partial \eta }}}
        \end{array}} \right){\text{d}}\xi {\text{d}}\eta

    此时面积微元的变换可以表示为：

    .. math::
        {\text{d}}S = \left\| {{\text{d}}\vec S} \right\| = \sqrt {\det {{\left( {\begin{array}{*{20}{c}}
          {\frac{{\partial x}}{{\partial \xi }}}&{\frac{{\partial y}}{{\partial \xi }}} \\
          {\frac{{\partial x}}{{\partial \eta }}}&{\frac{{\partial y}}{{\partial \eta }}}
        \end{array}} \right)}^2} + \det {{\left( {\begin{array}{*{20}{c}}
          {\frac{{\partial x}}{{\partial \xi }}}&{\frac{{\partial z}}{{\partial \xi }}} \\
          {\frac{{\partial x}}{{\partial \eta }}}&{\frac{{\partial z}}{{\partial \eta }}}
        \end{array}} \right)}^2} + \det {{\left( {\begin{array}{*{20}{c}}
          {\frac{{\partial y}}{{\partial \xi }}}&{\frac{{\partial z}}{{\partial \xi }}} \\
          {\frac{{\partial y}}{{\partial \eta }}}&{\frac{{\partial z}}{{\partial \eta }}}
        \end{array}} \right)}^2}} {\text{d}}\xi {\text{d}}\eta

    整理可得：

    .. math::
        {\text{d}}S = \sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}\frac{{\partial y}}{{\partial \eta }} - \frac{{\partial x}}{{\partial \eta }}\frac{{\partial y}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \xi }}\frac{{\partial x}}{{\partial \eta }} - \frac{{\partial z\partial x}}{{\partial \eta \partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}\frac{{\partial z}}{{\partial \eta }} - \frac{{\partial y}}{{\partial \eta }}\frac{{\partial z}}{{\partial \xi }}} \right)}^2}} {\text{d}}\xi {\text{d}}\eta

    同理:

    .. math::
        {\text{d}}S = \sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}\frac{{\partial y}}{{\partial \zeta }} - \frac{{\partial x}}{{\partial \zeta }}\frac{{\partial y}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \xi }}\frac{{\partial x}}{{\partial \zeta }} - \frac{{\partial z\partial x}}{{\partial \zeta \partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}\frac{{\partial z}}{{\partial \zeta }} - \frac{{\partial y}}{{\partial \zeta }}\frac{{\partial z}}{{\partial \xi }}} \right)}^2}} {\text{d}}\xi {\text{d}}\zeta

        {\text{d}}S = \sqrt {{{\left( {\frac{{\partial x}}{{\partial \zeta }}\frac{{\partial y}}{{\partial \eta }} - \frac{{\partial x}}{{\partial \eta }}\frac{{\partial y}}{{\partial \zeta }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \zeta }}\frac{{\partial x}}{{\partial \eta }} - \frac{{\partial z\partial x}}{{\partial \eta \partial \zeta }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \zeta }}\frac{{\partial z}}{{\partial \eta }} - \frac{{\partial y}}{{\partial \eta }}\frac{{\partial z}}{{\partial \zeta }}} \right)}^2}} {\text{d}}\zeta {\text{d}}\eta

    **3.1 以八节点六面体等参元为例子** ::

                     (-1,1,1)        (1,1,1)
                      7---------------6
                     /|              /|
                    / |     ζ  η    / |
                   /  |     | /    /  |
        (-1,-1,1) 4---+-----|/----5 (1,-1,1)
                  |   |     o--ξ  |   |
                  |   3-----------+---2 (1,1,-1)
                  |  /(-1,1,-1)   |  /
                  | /             | /
                  |/              |/
                  0---------------1
                 (-1,-1,-1)      (1,-1,-1)

    边界上的第一类曲面积分可以表示为：

    .. math::
        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\xi  =  \pm 1}} = \int_{ - 1}^1 {\int_{ - 1}^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}} } \sqrt {{{\left( {\frac{{\partial x}}{{\partial \zeta }}\frac{{\partial y}}{{\partial \eta }} - \frac{{\partial x}}{{\partial \eta }}\frac{{\partial y}}{{\partial \zeta }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \zeta }}\frac{{\partial x}}{{\partial \eta }} - \frac{{\partial z\partial x}}{{\partial \eta \partial \zeta }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \zeta }}\frac{{\partial z}}{{\partial \eta }} - \frac{{\partial y}}{{\partial \eta }}\frac{{\partial z}}{{\partial \zeta }}} \right)}^2}} {\text{d}}\zeta {\text{d}}\eta

        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\eta  =  \pm 1}} = \int_{ - 1}^1 {\int_{ - 1}^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}} } \sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}\frac{{\partial y}}{{\partial \zeta }} - \frac{{\partial x}}{{\partial \zeta }}\frac{{\partial y}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \xi }}\frac{{\partial x}}{{\partial \zeta }} - \frac{{\partial z\partial x}}{{\partial \zeta \partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}\frac{{\partial z}}{{\partial \zeta }} - \frac{{\partial y}}{{\partial \zeta }}\frac{{\partial z}}{{\partial \xi }}} \right)}^2}} {\text{d}}\xi {\text{d}}\zeta

        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\zeta  =  \pm 1}} = \int_{ - 1}^1 {\int_{ - 1}^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}} } \sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}\frac{{\partial y}}{{\partial \eta }} - \frac{{\partial x}}{{\partial \eta }}\frac{{\partial y}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \xi }}\frac{{\partial x}}{{\partial \eta }} - \frac{{\partial z\partial x}}{{\partial \eta \partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}\frac{{\partial z}}{{\partial \eta }} - \frac{{\partial y}}{{\partial \eta }}\frac{{\partial z}}{{\partial \xi }}} \right)}^2}} {\text{d}}\xi {\text{d}}\eta

    通过选取合适的积分点，采用数值积分即可求得积分值。

    **3.2 以四节点四面体等参元为例子** ::

       (0,0,1)
        3
        * **
        *   * *
        *     *  *
        *       *   2 (0,1,0)
        *        **  *
        ζ     *     * *
        |  η          **
        0---ξ * * * * * 1
       (0,0,0)         (1,0,0)

    边界上的第一类曲面积分可以表示为：

    .. math::
        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\xi  = 0}} = \int_0^1 {\int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}} } \sqrt {{{\left( {\frac{{\partial x}}{{\partial \zeta }}\frac{{\partial y}}{{\partial \eta }} - \frac{{\partial x}}{{\partial \eta }}\frac{{\partial y}}{{\partial \zeta }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \zeta }}\frac{{\partial x}}{{\partial \eta }} - \frac{{\partial z\partial x}}{{\partial \eta \partial \zeta }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \zeta }}\frac{{\partial z}}{{\partial \eta }} - \frac{{\partial y}}{{\partial \eta }}\frac{{\partial z}}{{\partial \zeta }}} \right)}^2}} {\text{d}}\zeta {\text{d}}\eta

        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\eta  = 0}} = \int_0^1 {\int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}} } \sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}\frac{{\partial y}}{{\partial \zeta }} - \frac{{\partial x}}{{\partial \zeta }}\frac{{\partial y}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \xi }}\frac{{\partial x}}{{\partial \zeta }} - \frac{{\partial z\partial x}}{{\partial \zeta \partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}\frac{{\partial z}}{{\partial \zeta }} - \frac{{\partial y}}{{\partial \zeta }}\frac{{\partial z}}{{\partial \xi }}} \right)}^2}} {\text{d}}\xi {\text{d}}\zeta

        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\zeta  = 0}} = \int_0^1 {\int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}} } \sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}\frac{{\partial y}}{{\partial \eta }} - \frac{{\partial x}}{{\partial \eta }}\frac{{\partial y}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \xi }}\frac{{\partial x}}{{\partial \eta }} - \frac{{\partial z\partial x}}{{\partial \eta \partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}\frac{{\partial z}}{{\partial \eta }} - \frac{{\partial y}}{{\partial \eta }}\frac{{\partial z}}{{\partial \xi }}} \right)}^2}} {\text{d}}\xi {\text{d}}\eta

    对于斜面 1-2-3，我们有平面方程 :math:`\zeta  = 1 - \xi - \eta` ：

    .. math::
        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\zeta  = 1 - \xi  - \eta }} = \int_0^1 {\int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}} } \sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}\frac{{\partial y}}{{\partial \eta }} - \frac{{\partial x}}{{\partial \eta }}\frac{{\partial y}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \xi }}\frac{{\partial x}}{{\partial \eta }} - \frac{{\partial z\partial x}}{{\partial \eta \partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}\frac{{\partial z}}{{\partial \eta }} - \frac{{\partial y}}{{\partial \eta }}\frac{{\partial z}}{{\partial \xi }}} \right)}^2}} \sqrt {1 + {{\left( {\frac{{\partial \zeta }}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial \zeta }}{{\partial \eta }}} \right)}^2}} {\text{d}}\xi {\text{d}}\eta

    所以有：

    .. math::
        {\left. {{{\mathbf{R}}^{\left( m \right)}}} \right|_{\zeta  = 1 - \xi  - \eta }} = \int_0^1 {\int_0^1 {{{\left( {{{\mathbf{N}}^{\left( m \right)}}} \right)}^{\text{T}}}{\mathbf{\bar p}}} } \sqrt {{{\left( {\frac{{\partial x}}{{\partial \xi }}\frac{{\partial y}}{{\partial \eta }} - \frac{{\partial x}}{{\partial \eta }}\frac{{\partial y}}{{\partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial z}}{{\partial \xi }}\frac{{\partial x}}{{\partial \eta }} - \frac{{\partial z\partial x}}{{\partial \eta \partial \xi }}} \right)}^2} + {{\left( {\frac{{\partial y}}{{\partial \xi }}\frac{{\partial z}}{{\partial \eta }} - \frac{{\partial y}}{{\partial \eta }}\frac{{\partial z}}{{\partial \xi }}} \right)}^2}} \sqrt 3 {\text{d}}\xi {\text{d}}\eta

    通过选取合适的积分点，采用数值积分即可求得积分值。
    """

    __slots__ = BaseBC.__slots__ + []

    def __init__(self, bc: BC, dof: Dof, mesh_data: MeshData, solver: Solver, amplitude: Optional[Amplitude]) -> None:
        super().__init__(bc, dof, mesh_data, solver, amplitude)
        self.create_dof_values()

    def get_surface_from_bc_element(self, bc_element_id: int, bc_element: ndarray) -> list[tuple[int, str]]:
        nodes = self.mesh_data.nodes
        elements = self.mesh_data.elements
        element_surface = []
        for element_id, element in enumerate(elements):
            is_element_surface = all(in1d(bc_element, element))
            if is_element_surface:
                nodes_in_element = in1d(element, bc_element)
                connectivity = elements[element_id]
                node_coords = nodes[connectivity]
                iso_element_type = get_iso_element_type(node_coords)
                iso_element_shape = iso_element_shape_dict[iso_element_type]
                surface_names = [surface_name for surface_name, nodes_on_surface in
                                 iso_element_shape.nodes_on_surface_dict.items() if
                                 all(nodes_on_surface == nodes_in_element)]
                if len(surface_names) == 1:
                    element_surface.append((element_id, surface_names[0]))
                else:
                    raise NotImplementedError(error_style(f'the surface of element {element_id} is wrong'))

        if len(element_surface) == 1:
            return element_surface
        else:
            raise NotImplementedError(error_style(f'the surface of bc_element {bc_element_id} is wrong'))

    def get_surface_from_elements_nodes(self, element_id: int, node_ids: list[int]) -> list[tuple[int, str]]:
        nodes = self.mesh_data.nodes
        elements = self.mesh_data.elements
        element_surface = []
        nodes_in_element = in1d(elements[element_id], node_ids)
        connectivity = elements[element_id]
        node_coords = nodes[connectivity]
        iso_element_type = get_iso_element_type(node_coords)
        iso_element_shape = iso_element_shape_dict[iso_element_type]
        surface_names = [surface_name for surface_name, nodes_on_surface in
                         iso_element_shape.nodes_on_surface_dict.items() if
                         sum(logical_and(nodes_in_element, nodes_on_surface)) == len(
                             iso_element_shape.bc_surface_nodes_dict[surface_name])]

        for surface_name in surface_names:
            element_surface.append((element_id, surface_name))

        if 1 <= len(element_surface) <= iso_element_shape.bc_surface_number:
            return element_surface
        else:
            raise NotImplementedError(error_style(f'the surface of element {element_id} is wrong'))

    def create_dof_values(self) -> None:
        dimension = self.mesh_data.dimension
        nodes = self.mesh_data.nodes
        elements = self.mesh_data.elements
        bc_elements = self.mesh_data.bc_elements

        node_sets = self.bc.node_sets
        element_sets = self.bc.element_sets
        bc_element_sets = self.bc.bc_element_sets
        bc_value = self.bc.value

        if bc_element_sets is not None:
            bc_element_ids = []
            for bc_element_set in bc_element_sets:
                bc_element_ids += list(self.mesh_data.bc_element_sets[bc_element_set])
            for bc_element_id in set(bc_element_ids):
                self.bc_surface += self.get_surface_from_bc_element(bc_element_id, bc_elements[bc_element_id])
        elif element_sets is not None and node_sets is not None:
            if element_sets == node_sets:
                element_ids = []
                for element_set in element_sets:
                    element_ids += list(self.mesh_data.element_sets[element_set])
                node_ids = []
                for node_set in node_sets:
                    node_ids += list(self.mesh_data.node_sets[node_set])
                for element_id in set(element_ids):
                    self.bc_surface += self.get_surface_from_elements_nodes(element_id, node_ids)
            else:
                raise NotImplementedError(
                    error_style(f'the name of element_sets {element_sets} and node_sets {node_sets} must be the same'))

        bc_dof_ids = []
        bc_fext = []
        bc_dof_names = self.bc.dof
        dof_names = self.dof.names

        for element_id, surface_name in self.bc_surface:
            connectivity = elements[element_id]
            node_coords = nodes[connectivity]
            iso_element_type = get_iso_element_type(node_coords)
            iso_element_shape = iso_element_shape_dict[iso_element_type]

            nodes_number = iso_element_shape.nodes_number
            bc_qp_weights = iso_element_shape.bc_qp_weights
            bc_qp_number = len(bc_qp_weights)
            bc_qp_shape_values = iso_element_shape.bc_qp_shape_values_dict[surface_name]
            bc_qp_shape_gradients = iso_element_shape.bc_qp_shape_gradients_dict[surface_name]
            bc_surface_coord = iso_element_shape.bc_surface_coord_dict[surface_name]
            surface_local_nodes = array(iso_element_shape.bc_surface_nodes_dict[surface_name])
            surface_nodes = elements[element_id][surface_local_nodes]
            surface_dof_ids = []
            for node_index in surface_nodes:
                for _, bc_dof_name in enumerate(bc_dof_names):
                    surface_dof_id = node_index * len(dof_names) + dof_names.index(bc_dof_name)
                    surface_dof_ids.append(surface_dof_id)

            bc_dof_ids += surface_dof_ids

            element_fext = zeros(nodes_number)

            for i in range(bc_qp_number):
                bc_qp_jacobi = dot(bc_qp_shape_gradients[i], node_coords).transpose()
                bc_qp_jacobi_sub = delete(bc_qp_jacobi, bc_surface_coord[0], axis=1)
                surface_weight = bc_surface_coord[3]
                if dimension == 2:
                    s = sum(bc_qp_jacobi_sub ** 2)
                    qp_fext = bc_qp_shape_values[i].transpose() * bc_qp_weights[i] * bc_value * sqrt(s) * surface_weight
                elif dimension == 3:
                    s = 0
                    for row in range(bc_qp_jacobi_sub.shape[0]):
                        s += det(delete(bc_qp_jacobi_sub, row, axis=0)) ** 2
                    qp_fext = bc_qp_shape_values[i].transpose() * bc_qp_weights[i] * bc_value * sqrt(s) * surface_weight
                else:
                    raise NotImplementedError(error_style(f'dimension {dimension} is not supported of the Neumann boundary condition'))

                element_fext += qp_fext

            surface_fext = []
            for fext in element_fext[surface_local_nodes]:
                for _ in range(len(bc_dof_names)):
                    surface_fext.append(fext)

            bc_fext += list(surface_fext)

        self.bc_dof_ids = array(bc_dof_ids)
        self.bc_fext = array(bc_fext)


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    # props = Properties()
    # props.read_file(r'..\..\..\examples\mechanical\plane\Job-1.toml')
    # bc_data = NeumannBCDistributed(props.bcs[2], props.dof, props.mesh_data, props.solver, None)
    # bc_data.show()

    # props = Properties()
    # props.read_file(r'..\..\..\examples\mechanical\quad8\Job-1.toml')
    # bc_data = NeumannBCDistributed(props.bcs[2], props.dof, props.mesh_data, props.solver, props.amplitudes[0])
    # bc_data.show()

    props = Properties()
    props.read_file(r'..\..\..\examples\mechanical\1element\tetra4\Job-1.toml')
    bc_data = NeumannBCDistributed(props.bcs[3], props.dof, props.mesh_data, props.solver, None)
    bc_data.show()



# -*- coding: utf-8 -*-
"""

"""
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style


class PhaseFieldDamage(BaseMaterial):
    """
    相场断裂材料。

    支持的截面属性：('Volume', 'PlaneStress', 'PlaneStrain')

    :ivar gc: surface energy to create a unit fracture surface gc
    :vartype gc: float

    :ivar lc: length scale parameter to measure the damage diffusion lc
    :vartype lc: float
    """

    __slots_dict__: dict = {
        'gc': ('float', 'surface energy to create a unit fracture surface gc'),
        'lc': ('float', 'length scale parameter to measure the damage diffusion lc'),
    }

    __slots__ = BaseMaterial.__slots__ + [slot for slot in __slots_dict__.keys()]

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('Volume', 'PlaneStress', 'PlaneStrain')

        self.data_keys = ['surface energy to create a unit fracture surface gc',
                          'length scale parameter to measure the damage diffusion lc']

        if len(self.material.data) != len(self.data_keys):
            raise NotImplementedError(error_style(self.get_data_length_error_msg()))
        else:
            for i, key in enumerate(self.data_keys):
                self.data_dict[key] = material.data[i]

        self.gc: float = self.data_dict['surface energy to create a unit fracture surface gc']
        self.lc: float = self.data_dict['length scale parameter to measure the damage diffusion lc']

        self.create_tangent()


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(PhaseFieldDamage.__slots_dict__)

    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'..\..\..\examples\mechanical_phase\rectangle\Job-1.toml')
    material_data = PhaseFieldDamage(props.materials[1], 3, props.sections[0])
    material_data.show()



# -*- coding: utf-8 -*-
"""

"""
from pyfem.io.BaseIO import BaseIO


class Parameter(BaseIO):
    """
    定义需要读取的参数信息。

    :ivar file: 参数文件路径
    :vartype file: str
    """

    __slots_dict__: dict = {
        'file': ('str', '参数文件路径')
    }

    __slots__: list = [slot for slot in __slots_dict__.keys()]

    def __init__(self) -> None:
        super().__init__()
        self.file: str = None  # type: ignore


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(Parameter.__slots_dict__)

    mesh = Parameter()
    mesh.show()



# -*- coding: utf-8 -*-
"""

"""
from copy import deepcopy

from numpy import zeros
from numpy.linalg import norm
from scipy.sparse.linalg import splu  # type: ignore

from pyfem.assembly.Assembly import Assembly
from pyfem.fem.constants import DTYPE
from pyfem.io.Solver import Solver
from pyfem.io.write_vtk import write_vtk, write_pvd
from pyfem.solvers.BaseSolver import BaseSolver
from pyfem.utils.colors import info_style, warn_style, error_style
from pyfem.utils.logger import logger, logger_sta


class NonlinearSolver(BaseSolver):
    r"""
    非线性求解器。

    :ivar PENALTY: 罚系数
    :vartype PENALTY: float

    :ivar FORCE_TOL: 残差容限
    :vartype FORCE_TOL: float

    :ivar MAX_NITER: 最大迭代次数
    :vartype MAX_NITER: int

    对于准静态过程，需要求解的线性系统可以表示为：

    .. math::
        {{\mathbf{f}}_{{\text{ext}}}} - {{\mathbf{f}}_{{\text{int}}}} = {\mathbf{0}}

    虽然在静态力学过程中，时间不再起作用，但是我们仍然需要一个参数来排列事件的顺序。出于这个原因，我们将继续在静态力学过程中使用“时间”的概念来表示加载顺序。特别地，时间的概念可以用来将完整的外部载荷分解为若干个增量加载步骤。这样做的原因如下：

    1. 从非线性连续体模型的离散化得到的代数方程组是非线性的，因此需要使用迭代过程来求解。对于非常大的加载步骤（例如在一个增量步中施加整个载荷），通常很难获得一个正确收敛的解。对于大多数常用的迭代过程，包括牛顿-拉夫逊方法，其收敛半径都是有限的。

    2. 实验证明，大多数材料表现出路径相关的行为。这意味着根据所遵循的应变路径获得的应力值是不同的。例如，当我们首先对一个平板施加拉伸应变增量，然后施加剪切应变增量时，所得到的应力可能会不同，或者当以相反的顺序施加相同的应变增量时，结果应力也可能不同。显然，只有在应变增量相对较小的情况下，才能正确预测结构行为，以便尽可能地按照应变路径进行。

    参考文献：

    [1] Non‐Linear Finite Element Analysis of Solids and Structures, John Wiley & Sons, Ltd, 2012, 31-62, https://doi.org/10.1002/9781118375938.ch2
    """

    __slots_dict__: dict = {
        'PENALTY': ('float', '罚系数'),
        'FORCE_TOL': ('float', '残差容限'),
        'MAX_NITER': ('int', '最大迭代次数')
    }

    __slots__ = BaseSolver.__slots__ + [slot for slot in __slots_dict__.keys()]

    def __init__(self, assembly: Assembly, solver: Solver) -> None:
        super().__init__()
        self.assembly = assembly
        self.solver = solver
        self.dof_solution = zeros(self.assembly.total_dof_number)
        self.PENALTY: float = 1.0e128
        self.FORCE_TOL: float = 1.0e-3
        self.MAX_NITER: int = 16

    def run(self) -> int:
        if self.assembly.props.solver.option in [None, '', 'NR', 'NewtonRaphson']:
            return self.incremental_iterative_solve('NR')
        elif self.assembly.props.solver.option in ['IT', 'InitialTangent']:
            return self.incremental_iterative_solve('IT')
        else:
            raise NotImplementedError(error_style(
                f'unsupported option \'{self.assembly.props.solver.option}\' of {self.assembly.props.solver.type}'))

    def incremental_iterative_solve(self, option: str) -> int:
        timer = self.assembly.timer
        timer.total_time = self.solver.total_time
        timer.dtime = self.solver.initial_dtime
        timer.time0 = self.solver.start_time
        timer.increment = 0
        timer.frame_ids.append(0)

        self.assembly.update_element_field_variables()
        self.assembly.assembly_field_variables()
        for output in self.assembly.props.outputs:
            if output.is_save:
                if output.type == 'vtk':
                    write_vtk(self.assembly)

        x = []
        y = []
        t = []
        x.append(self.assembly.element_data_list[0].element_average_field_variables['E11'])
        y.append(self.assembly.element_data_list[0].element_average_field_variables['S11'])
        t.append(0.0)

        increment: int = 1
        attempt: int = 1

        for i in range(1, self.solver.max_increment):

            timer.time1 = timer.time0 + timer.dtime
            timer.increment = increment

            logger.info(f'increment = {increment}, attempt = {attempt}, time = {timer.time1:14.9f}, dtime = {timer.dtime:14.9f}')

            self.assembly.ddof_solution = zeros(self.assembly.total_dof_number, dtype=DTYPE)
            self.assembly.update_element_data()

            is_convergence = False

            for niter in range(self.MAX_NITER):
                self.assembly.assembly_global_stiffness()
                fint = self.assembly.fint
                rhs = deepcopy(self.assembly.fext)
                if niter == 0:
                    for bc_data in self.assembly.bc_data_list:
                        amplitude_increment = bc_data.get_amplitude(timer.time1) - bc_data.get_amplitude(timer.time0)
                        if bc_data.bc.category == 'DirichletBC':
                            for bc_dof_id, bc_dof_value in zip(bc_data.bc_dof_ids, bc_data.bc_dof_values):
                                self.assembly.global_stiffness[bc_dof_id, bc_dof_id] += self.PENALTY
                                rhs[bc_dof_id] += bc_dof_value * self.PENALTY * amplitude_increment
                        elif bc_data.bc.category == 'NeumannBC':
                            for bc_dof_id, bc_fext in zip(bc_data.bc_dof_ids, bc_data.bc_fext):
                                rhs[bc_dof_id] += bc_fext * amplitude_increment
                                self.assembly.fext[bc_dof_id] += bc_fext * amplitude_increment
                else:
                    for bc_data in self.assembly.bc_data_list:
                        if bc_data.bc.category == 'DirichletBC':
                            for bc_dof_id, bc_dof_value in zip(bc_data.bc_dof_ids, bc_data.bc_dof_values):
                                self.assembly.global_stiffness[bc_dof_id, bc_dof_id] += self.PENALTY
                                rhs[bc_dof_id] = 0.0 * self.PENALTY

                try:
                    LU = splu(self.assembly.global_stiffness)
                    da = LU.solve(rhs - fint)
                except RuntimeError as e:
                    is_convergence = False
                    print(error_style(f"Catch RuntimeError exception: {e}"))
                    break

                self.assembly.ddof_solution += da
                if option == 'NR':
                    self.assembly.update_element_data()
                elif option == 'IT':
                    self.assembly.update_element_data_without_stiffness()
                self.assembly.assembly_fint()

                f_residual = self.assembly.fext - self.assembly.fint
                f_residual[self.assembly.bc_dof_ids] = 0
                if norm(self.assembly.fext) < 1.0e-16:
                    f_residual = norm(f_residual)
                else:
                    f_residual = norm(f_residual) / norm(self.assembly.fext)
                # f_residual = max(abs(f_residual))

                logger.log(21, f'  niter = {niter}, residual = {f_residual}')

                if timer.is_reduce_dtime:
                    timer.is_reduce_dtime = False
                    is_convergence = False
                    break

                if f_residual < self.FORCE_TOL:
                    is_convergence = True
                    break

            if is_convergence:
                logger.info(f'  increment {increment} is convergence')
                logger_sta.info(f'{1:4}  {increment:9}  {attempt:3}  {0:6}  {niter:5}  {niter:5}  {timer.time1:14.6f}  {timer.time1:14.6f}  {timer.dtime:14.6f}')

                self.assembly.update_element_data()
                self.assembly.dof_solution += self.assembly.ddof_solution
                # 调换了上面两行代码的顺序，基于t时刻的自由度值及t+dt时刻的自由度增量值对单元信息进行更新，之后在将所有单元的自由度值更新为t+dt时刻。

                self.assembly.update_element_state_variables()
                self.assembly.update_element_field_variables()
                self.assembly.assembly_field_variables()

                for output in self.assembly.props.outputs:
                    if output.is_save:
                        if output.type == 'vtk':
                            write_vtk(self.assembly)

                timer.time0 = timer.time1
                timer.frame_ids.append(increment)
                increment += 1
                attempt = 1
                timer.dtime *= 1.1
                if timer.dtime >= self.solver.max_dtime:
                    timer.dtime = self.solver.max_dtime
                if timer.time0 + timer.dtime >= self.solver.total_time:
                    timer.dtime = self.solver.total_time - timer.time0

                x.append(self.assembly.element_data_list[0].element_average_field_variables['E11'])
                y.append(self.assembly.element_data_list[0].element_average_field_variables['S11'])
                t.append(timer.time0)

            else:
                attempt += 1
                timer.dtime *= 0.5
                logger.warning(f'  increment {increment} is divergence, dtime is reduced to {timer.dtime}')

                if timer.dtime <= self.assembly.props.solver.min_dtime:
                    for output in self.assembly.props.outputs:
                        if output.is_save:
                            if output.type == 'vtk':
                                write_pvd(self.assembly)
                    logger.error(f'Computation is ended with error: the dtime {timer.dtime} is less than the minimum value')
                    return -1

                self.assembly.ddof_solution = zeros(self.assembly.total_dof_number, dtype=DTYPE)
                self.assembly.goback_element_state_variables()
                self.assembly.update_element_data()
                self.assembly.assembly_fint()

            if timer.is_done():
                for output in self.assembly.props.outputs:
                    if output.is_save:
                        if output.type == 'vtk':
                            write_pvd(self.assembly)
                break

        if not timer.is_done():
            logger.error('maximum increment is reached')
            return -1
        else:
            return 0


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(NonlinearSolver.__slots_dict__)

    from pyfem.Job import Job

    job = Job(r'..\..\..\examples\mechanical\beam\Job-1.toml')
    solver = NonlinearSolver(job.assembly, job.props.solver)
    solver.show()



# -*- coding: utf-8 -*-
"""

"""
from numpy import zeros, ndarray, array, sum, dot, maximum
from numpy.linalg import inv, norm

from pyfem.fem.constants import DTYPE
from pyfem.utils.colors import error_style


def inverse(qp_jacobis: ndarray, qp_jacobi_dets: ndarray) -> ndarray:
    r"""
    **求逆矩阵**

    :param qp_jacobis: 积分点处的雅克比矩阵列表
    :type qp_jacobis: ndarray

    :param qp_jacobi_dets: 积分点处的雅克比矩阵行列式列表
    :type qp_jacobi_dets: ndarray

    :return: 逆矩阵列表
    :rtype: ndarray

    当输入为 2×2 和 3×3 的矩阵直接通过解析式计算，其余的情况返回 :py:meth:`numpy.linalg.inv()` 函数的计算结果。

    对于 2×2 的矩阵：

    .. math::
        {\mathbf{A}} = \left[ {\begin{array}{*{20}{c}}
          {{A_{11}}}&{{A_{12}}} \\
          {{A_{21}}}&{{A_{22}}}
        \end{array}} \right]

    .. math::
        {{\mathbf{A}}^{ - 1}} = \frac{1}{{\det \left( {\mathbf{A}} \right)}}\left[ {\begin{array}{*{20}{c}}
          {{A_{22}}}&{ - {A_{12}}} \\
          { - {A_{21}}}&{{A_{11}}}
        \end{array}} \right]

    对于 3×3 的矩阵：

    .. math::
        {\mathbf{A}} = \left[ {\begin{array}{*{20}{c}}
          {{A_{11}}}&{{A_{12}}}&{{A_{13}}} \\
          {{A_{21}}}&{{A_{22}}}&{{A_{23}}} \\
          {{A_{31}}}&{{A_{32}}}&{{A_{33}}}
        \end{array}} \right]

    .. math::
        {{\mathbf{A}}^{ - 1}} = \frac{1}{{\det \left( {\mathbf{A}} \right)}}\left[ {\begin{array}{*{20}{c}}
          {{A_{22}}{A_{33}} - {A_{23}}{A_{32}}}&{{A_{13}}{A_{32}} - {A_{12}}{A_{33}}}&{{A_{12}}{A_{23}} - {A_{13}}{A_{22}}} \\
          {{A_{23}}{A_{31}} - {A_{21}}{A_{33}}}&{{A_{11}}{A_{33}} - {A_{13}}{A_{31}}}&{{A_{13}}{A_{21}} - {A_{11}}{A_{23}}} \\
          {{A_{21}}{A_{32}} - {A_{22}}{A_{31}}}&{{A_{12}}{A_{31}} - {A_{11}}{A_{32}}}&{{A_{11}}{A_{22}} - {A_{12}}{A_{21}}}
        \end{array}} \right]
    """

    qp_jacobi_invs = []
    for A, det_A in zip(qp_jacobis, qp_jacobi_dets):
        if A.shape == (2, 2):
            qp_jacobi_invs.append(array([[A[1][1], -A[0][1]], [-A[1][0], A[0][0]]]) / det_A)
        elif A.shape == (3, 3):
            qp_jacobi_invs.append(array([[(A[1][1] * A[2][2] - A[1][2] * A[2][1]),
                                          (A[0][2] * A[2][1] - A[0][1] * A[2][2]),
                                          (A[0][1] * A[1][2] - A[0][2] * A[1][1])],
                                         [(A[1][2] * A[2][0] - A[1][0] * A[2][2]),
                                          (A[0][0] * A[2][2] - A[0][2] * A[2][0]),
                                          (A[0][2] * A[1][0] - A[0][0] * A[1][2])],
                                         [(A[1][0] * A[2][1] - A[1][1] * A[2][0]),
                                          (A[0][1] * A[2][0] - A[0][0] * A[2][1]),
                                          (A[0][0] * A[1][1] - A[0][1] * A[1][0])]]) / det_A)
        else:
            return inv(qp_jacobis)
    return array(qp_jacobi_invs)


def get_transformation(u: ndarray, v: ndarray, w: ndarray,
                       u_prime: ndarray, v_prime: ndarray, w_prime: ndarray) -> ndarray:
    r"""
    **计算空间变换矩阵**

    :param u: :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 坐标系下的1号矢量
    :type u: ndarray

    :param v: :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 坐标系下的2号矢量
    :type v: ndarray

    :param w: :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 坐标系下的3号矢量
    :type w: ndarray

    :param u_prime: :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 坐标系下的1号矢量
    :type u_prime: ndarray

    :param v_prime: :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 坐标系下的2号矢量
    :type v_prime: ndarray

    :param w_prime: :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 坐标系下的3号矢量
    :type w_prime: ndarray

    :return: 空间变换矩阵（线性）
    :rtype: ndarray

    设 :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 和 :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 是空间 :math:`{{\mathbb{R}}^3}` 的两组基。
    如果矩阵 :math:`\mathbf{T}` 描述了两组基对应的线性变换 :math:`{{\mathbb{R}}^3} \Rightarrow {{\mathbb{R}}^3}`，则对于空间中的三个任意矢量有：

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]\left\{ {\begin{array}{*{20}{c}}
          {{u_1}} \\
          {{u_2}} \\
          {{u_3}}
        \end{array}} \right\} = \left\{ {\begin{array}{*{20}{c}}
          {{{u'}_1}} \\
          {{{u'}_2}} \\
          {{{u'}_3}}
        \end{array}} \right\}

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]\left\{ {\begin{array}{*{20}{c}}
          {{v_1}} \\
          {{v_2}} \\
          {{v_3}}
        \end{array}} \right\} = \left\{ {\begin{array}{*{20}{c}}
          {{{v'}_1}} \\
          {{{v'}_2}} \\
          {{{v'}_3}}
        \end{array}} \right\}

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]\left\{ {\begin{array}{*{20}{c}}
          {{w_1}} \\
          {{w_2}} \\
          {{w_3}}
        \end{array}} \right\} = \left\{ {\begin{array}{*{20}{c}}
          {{{w'}_1}} \\
          {{{w'}_2}} \\
          {{{w'}_3}}
        \end{array}} \right\}

    合并可得：

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]\left[ {\begin{array}{*{20}{c}}
          {{u_1}}&{{v_1}}&{{w_1}} \\
          {{u_2}}&{{v_2}}&{{w_2}} \\
          {{u_3}}&{{v_3}}&{{w_3}}
        \end{array}} \right] = \left[ {\begin{array}{*{20}{c}}
          {{{u'}_1}}&{{{v'}_1}}&{{{w'}_1}} \\
          {{{u'}_2}}&{{{v'}_2}}&{{{w'}_2}} \\
          {{{u'}_3}}&{{{v'}_3}}&{{{w'}_3}}
        \end{array}} \right]

    因此变换矩阵 :math:`\mathbf{T}` 可以表示为：

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right] = \left[ {\begin{array}{*{20}{c}}
          {{{u'}_1}}&{{{v'}_1}}&{{{w'}_1}} \\
          {{{u'}_2}}&{{{v'}_2}}&{{{w'}_2}} \\
          {{{u'}_3}}&{{{v'}_3}}&{{{w'}_3}}
        \end{array}} \right]{\left[ {\begin{array}{*{20}{c}}
          {{u_1}}&{{v_1}}&{{w_1}} \\
          {{u_2}}&{{v_2}}&{{w_2}} \\
          {{u_3}}&{{v_3}}&{{w_3}}
        \end{array}} \right]^{ - 1}}

    记为：

    .. math::
        {\mathbf{T}} = {\mathbf{A'}} \cdot {\left( {\mathbf{A}} \right)^{ - 1}}

    特殊的，对于两个空间直角坐标系的映射，即 :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 和 :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 均为单位正交基，且变换需要满足  :math:`{{\mathbb{R}}^3} \Rightarrow {{\mathbb{R}}^3}` ，则 :math:`{\mathbf{A}}` 和  :math:`{{\mathbf{A'}}}` 必须满秩且为正交矩阵。
    """

    A = array([u, v, w])
    A_prime = array([u_prime, v_prime, w_prime])
    T = dot(A_prime, inv(A))
    return T


def vogit_array_to_tensor(vogit_array: ndarray, dimension: int) -> ndarray:
    r"""
    **Voigt记法数组转换为2阶张量**

    :param vogit_array: Voigt记法数组
    :type vogit_array: ndarray

    :param dimension: 空间维度
    :type dimension: int

    :return: 2阶张量
    :rtype: ndarray

    映射方式：

    .. math::
        \left\{ {\begin{array}{*{20}{c}}
          {{T_{11}}} \\
          {{T_{22}}} \\
          {{T_{33}}} \\
          {{T_{12}}} \\
          {{T_{13}}} \\
          {{T_{23}}}
        \end{array}} \right\} \to \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]
    """

    tensor = zeros(shape=(dimension, dimension))
    if dimension == 2:
        tensor[0, 0] = vogit_array[0]
        tensor[1, 1] = vogit_array[1]
        tensor[0, 1] = vogit_array[2]
        tensor[1, 0] = vogit_array[2]
    elif dimension == 3:
        tensor[0, 0] = vogit_array[0]
        tensor[1, 1] = vogit_array[1]
        tensor[2, 2] = vogit_array[2]
        tensor[0, 1] = vogit_array[3]
        tensor[0, 2] = vogit_array[4]
        tensor[1, 2] = vogit_array[5]
        tensor[1, 0] = vogit_array[3]
        tensor[2, 0] = vogit_array[4]
        tensor[2, 1] = vogit_array[5]
    else:
        raise NotImplementedError(error_style(f'unsupported dimension {dimension}'))
    return tensor


def get_voigt_transformation(transformation: ndarray) -> ndarray:
    """
    **获取旋转矩阵的Voigt形式**

    :param transformation: Voigt记法数组
    :type transformation: ndarray

    :return: voigt_transformation
    :rtype: ndarray
    """

    voigt_transformation = zeros(shape=(6, 6), dtype=DTYPE)

    a11 = transformation[0][0]
    a12 = transformation[0][1]
    a13 = transformation[0][2]
    a21 = transformation[1][0]
    a22 = transformation[1][1]
    a23 = transformation[1][2]
    a31 = transformation[2][0]
    a32 = transformation[2][1]
    a33 = transformation[2][2]

    voigt_transformation[0][0] = a11 ** 2
    voigt_transformation[0][1] = a12 ** 2
    voigt_transformation[0][2] = a13 ** 2
    voigt_transformation[0][3] = 2 * a11 * a12
    voigt_transformation[0][4] = 2 * a11 * a13
    voigt_transformation[0][5] = 2 * a12 * a13

    voigt_transformation[1][0] = a21 ** 2
    voigt_transformation[1][1] = a22 ** 2
    voigt_transformation[1][2] = a23 ** 2
    voigt_transformation[1][3] = 2 * a21 * a22
    voigt_transformation[1][4] = 2 * a21 * a23
    voigt_transformation[1][5] = 2 * a22 * a23

    voigt_transformation[2][0] = a31 ** 2
    voigt_transformation[2][1] = a32 ** 2
    voigt_transformation[2][2] = a33 ** 2
    voigt_transformation[2][3] = 2 * a31 * a32
    voigt_transformation[2][4] = 2 * a31 * a33
    voigt_transformation[2][5] = 2 * a32 * a33

    voigt_transformation[3][0] = a11 * a21
    voigt_transformation[3][1] = a12 * a22
    voigt_transformation[3][2] = a13 * a23
    voigt_transformation[3][3] = a12 * a21 + a11 * a22
    voigt_transformation[3][4] = a13 * a21 + a11 * a23
    voigt_transformation[3][5] = a13 * a22 + a12 * a23

    voigt_transformation[4][0] = a11 * a31
    voigt_transformation[4][1] = a12 * a32
    voigt_transformation[4][2] = a13 * a33
    voigt_transformation[4][3] = a12 * a31 + a11 * a32
    voigt_transformation[4][4] = a13 * a31 + a11 * a33
    voigt_transformation[4][5] = a13 * a32 + a12 * a33

    voigt_transformation[5][0] = a21 * a31
    voigt_transformation[5][1] = a22 * a32
    voigt_transformation[5][2] = a23 * a33
    voigt_transformation[5][3] = a22 * a31 + a21 * a32
    voigt_transformation[5][4] = a23 * a31 + a21 * a33
    voigt_transformation[5][5] = a23 * a32 + a22 * a33

    return voigt_transformation


def get_decompose_energy(strain: ndarray, stress: ndarray, dimension: int):
    # strain = array_to_tensor_order_2(strain, dimension)
    # # stress = array_to_tensor_order_2(stress, dimension)
    #
    # principle_strain_value, principle_strain_vector = eig(strain)
    # # principle_stress_value, principle_stress_vector = eig(stress)
    #
    # strain_positive = zeros(shape=(dimension, dimension))
    # strain_negative = zeros(shape=(dimension, dimension))
    #
    # # stress_positive = zeros(shape=(dimension, dimension))
    # # stress_negative = zeros(shape=(dimension, dimension))
    #
    # for i in range(dimension):
    #     strain_positive += 0.5 * (principle_strain_value[i] + abs(principle_strain_value[i])) * \
    #                        tensordot(principle_strain_vector[:, i], principle_strain_vector[:, i], 0)
    #
    #     strain_negative += 0.5 * (principle_strain_value[i] - abs(principle_strain_value[i])) * \
    #                        tensordot(principle_strain_vector[:, i], principle_strain_vector[:, i], 0)
    #
    #     # stress_positive += 0.5 * (principle_stress_value[i] + abs(principle_stress_value[i])) * \
    #     #                    tensordot(principle_stress_vector[:, i], principle_stress_vector[:, i], 0)
    #     #
    #     # stress_negative += 0.5 * (principle_stress_value[i] - abs(principle_stress_value[i])) * \
    #     #                    tensordot(principle_stress_vector[:, i], principle_stress_vector[:, i], 0)
    #
    # E = 1.0e5
    # nu = 0.25
    #
    # mu = E / (2 * (1 + nu))
    # lame = (E * nu) / ((1 + nu) * (1 - 2 * nu))
    #
    # energy_positive = 0.5 * lame * (0.5 * (strain.trace() + abs(strain.trace()))) ** 2 + \
    #                   mu * (strain_positive * strain_positive).trace()
    #
    # energy_negative = 0.5 * lame * (0.5 * (strain.trace() - abs(strain.trace()))) ** 2 + \
    #                   mu * (strain_negative * strain_negative).trace()

    energy_positive = 0.5 * sum(stress * strain)

    return energy_positive, energy_positive


def operations_for_symtensor_antisymtensor(sym_tensor: ndarray, antisym_tensor: ndarray) -> ndarray:
    r"""
    **获取反对称张量乘以对称张量减去对称张量乘以反对称张量的结果**

    :param sym_tensor: :math:`\left[ {{a_{11}}{\text{ }}{a_{22}}{\text{ }}{a_{33}}{\text{ }}{a_{12}}{\text{ }}{a_{13}}{\text{ }}{a_{23}}} \right]` 采用 Vogit 记法的对称张量数组
    :type sym_tensor: ndarray

    :param antisym_tensor: :math:`\left[ {0{\text{ }}0{\text{ }}0{\text{ }}{b_{12}}{\text{ }}{b_{13}}{\text{ }}{b_{23}}} \right]` 采用类似 Vogit 记法的反对称张量数组
    :type antisym_tensor: ndarray

    :return: 反对称张量乘以对称张量减去对称张量乘以反对称张量的结果
    :rtype: ndarray

    首先分别将对称张量数组 :math:`\left[ {{a_{11}}{\text{ }}{a_{22}}{\text{ }}{a_{33}}{\text{ }}{a_{12}}{\text{ }}{a_{13}}{\text{ }}{a_{23}}} \right]` 和
    反对称张量数组 :math:`\left[ {0{\text{ }}0{\text{ }}0{\text{ }}{b_{12}}{\text{ }}{b_{13}}{\text{ }}{b_{23}}} \right]` 写成矩阵形式：

    .. math::
        A_{sym} = \left[ {\begin{array}{*{20}{c}}
          {{a_{11}}}&{{a_{12}}}&{{a_{13}}} \\
          {{a_{12}}}&{{a_{22}}}&{{a_{23}}} \\
          {{a_{13}}}&{{a_{23}}}&{{a_{33}}}
        \end{array}} \right],B_{antisym} = \left[ {\begin{array}{*{20}{c}}
          0&{{b_{12}}}&{{b_{13}}} \\
          { - {b_{12}}}&0&{{b_{23}}} \\
          { - {b_{13}}}&{ - {b_{23}}}&0
        \end{array}} \right]

    计算结果 :math:`{B_{antisym}}{A_{sym}} - {A_{sym}}{B_{antisym}}` 表示为：

    .. math::
        \begin{gathered}
          {B_{antisym}}{A_{sym}} - {A_{sym}}{B_{antisym}} = \left[ {\begin{array}{*{20}{c}}
          0&0&0&{2{b_{12}}}&{2{b_{13}}}&0 \\
          0&0&0&{ - 2{b_{12}}}&0&{2{b_{23}}} \\
          0&0&0&0&{ - 2{b_{13}}}&{ - 2{b_{23}}} \\
          { - {b_{12}}}&{{b_{12}}}&0&0&{{b_{23}}}&{{b_{13}}} \\
          { - {b_{13}}}&0&{{b_{13}}}&{ - {b_{23}}}&0&{{b_{12}}} \\
          0&{ - {b_{23}}}&{{b_{23}}}&{ - {b_{13}}}&{ - {b_{12}}}&0
        \end{array}} \right]\left[ \begin{gathered}
          {a_{11}} \hfill \\
          {a_{22}} \hfill \\
          {a_{33}} \hfill \\
          {a_{12}} \hfill \\
          {a_{13}} \hfill \\
          {a_{23}} \hfill \\
        \end{gathered}  \right] \hfill \\
           = \left[ \begin{gathered}
          2{b_{12}}{a_{12}} + 2{b_{13}}{a_{13}} \hfill \\
           - 2{b_{12}}{a_{12}} + 2{b_{23}}{a_{23}} \hfill \\
           - 2{b_{13}}{a_{13}} - 2{b_{23}}{a_{23}} \hfill \\
           - {b_{12}}{a_{11}} + {b_{12}}{a_{22}} + {b_{23}}{a_{13}} + {b_{13}}{a_{23}} \hfill \\
           - {b_{13}}{a_{11}} + {b_{13}}{a_{33}} - {b_{23}}{a_{12}} + {b_{12}}{a_{23}} \hfill \\
           - {b_{23}}{a_{22}} + {b_{23}}{a_{33}} - {b_{13}}{a_{12}} - {b_{12}}{a_{13}} \hfill \\
        \end{gathered}  \right] = T{A_{sym}} \hfill \\
        \end{gathered}
    """
    result = zeros(shape=(len(antisym_tensor), 6), dtype=DTYPE)

    for i in range(len(antisym_tensor)):
        T = array([[0, 0, 0, 2 * antisym_tensor[i, 3], 2 * antisym_tensor[i, 4], 0],
                   [0, 0, 0, -2 * antisym_tensor[i, 3], 0, 2 * antisym_tensor[i, 5]],
                   [0, 0, 0, 0, -2 * antisym_tensor[i, 4], -2 * antisym_tensor[i, 5]],
                   [-antisym_tensor[i, 3], antisym_tensor[i, 3], 0, 0, antisym_tensor[i, 5], antisym_tensor[i, 4]],
                   [-antisym_tensor[i, 4], 0, antisym_tensor[i, 4], -antisym_tensor[i, 5], 0, antisym_tensor[i, 3]],
                   [0, -antisym_tensor[i, 5], antisym_tensor[i, 5], -antisym_tensor[i, 4], -antisym_tensor[i, 3], 0]])
        result[i, :] = dot(T, sym_tensor)

    return result


if __name__ == '__main__':
    u = array([1, 0, 0])
    v = array([0, 1, 0])
    w = array([0, 0, 1])

    u_prime = array([1, 0, 0])
    v_prime = array([0, 1, 0])
    w_prime = array([0, 0, 0.6187])
    T = get_transformation(u, v, w, u_prime, v_prime, w_prime)

    n = array([1, 0, 0])
    n = dot(T, n)
    n *= 1.0 / norm(n)

    print(n)
